﻿namespace Toast

import Toast.FreeSL
import Mogre
import System.Math
import System.Windows.Forms
import System.Threading
import System.IO

enum SoundDistanceModel:
    Inverse = 0xD001
    ClampedInverse
    Linear
    ClampedLinear
    Exponent
    ClampedExponent

enum SoundState:
    Stopped
    Playing
    Paused
    
enum PlayLimitMode:
    ReplaceOld
    RejectNew
    
enum SoundEnvironment:
    Generic
    PaddedCell
    Room
    Bathroom
    LivingRoom
    StoneRoom
    Auditorium
    ConcertHall
    Cave
    Arena
    Hangar
    CarpetedHallway
    Hallway
    StoneCorridor
    Alley
    Forest
    City
    Mountains
    Quarry
    Plain
    ParkingLot
    SewerPipe
    Underwater
    Drugged
    Dizzy
    Psychotic
    None
    
internal private class SoundLoader(ContentLoader):
    
    override def _loadFromFolder(fileName as string, groupName as string):
        
        if FileInfo(fileName).Length > Audio.MaxLoadSize * 1024:
            soundObj = fslStreamSound(fileName)
            streaming = true
        else:
            soundObj = fslLoadSound(fileName)
            streaming = false
            
        name = fileName
        if name.Contains('\\'): name = name[name.LastIndexOf('\\') + 1:]
        Audio._addToLoaded(name.ToLower(), fileName, '', streaming,
                           soundObj)
        return soundObj
    
    override def _loadFromZip(fileName as string, zipName as string,
                              groupName as string):
        
        soundObj = fslLoadSoundFromZip(zipName, fileName)
        Audio._addToLoaded(fileName.ToLower(), fileName, zipName, false,
                           soundObj)
        return soundObj
        
    override def _unload(content):
        
        fslFreeSound(cast(uint, content), true)
        Audio._removeFromLoaded(content)
    
internal private class VolumeFader:
    
    _fadee = null
    _fadeTime = 0.0f
    _fading = false
    _v1 = 0.0f
    _v2 = 0.0f
    _duration = 0.0f
    _callback as callable(Sound)
    
    IsFading:
        get: return _fading
    
    def constructor(fadee as Sound):
        
        _fadee = fadee
        
    def constructor(fadee as SoundGroup):
        
        _fadee = fadee
    
    def _fade(v1 as single, v2 as single, duration as single,
              callback as callable(Sound)):
        
        _v1 = v1
        _v2 = v2
        _duration = duration
        _fadeTime = 0.0f
        _callback = callback
        _fading = true
        
    def _cancelFading():
        
        _fadeTime = 0
        _fading = false
        
    def _update():
        
        if _fadeTime >= _duration:
            if _fadee isa Sound: (_fadee as Sound).Volume = _v2
            else: (_fadee as SoundGroup).Volume = _v2
            if _callback: _callback(_fadee)
            _fadeTime = 0
            _callback = null
            _fading = false
        else:
            v = _v1 + (_v2 - _v1) * _fadeTime / _duration
            if _fadee isa Sound: (_fadee as Sound).Volume = v
            else: (_fadee as SoundGroup).Volume = v
            _fadeTime += Core.ActualFrameTime
        
abstract class Sound:
    
    _name as string
    _state as SoundState
    _fader as VolumeFader
    _soundObj as uint
    _volume = 1.0f
    _pitch = 1.0f
    _groups = []
    _muted = false
    _stream as bool
    _async = []
    
    [Property(Importance)] _importance = 1.0f
    Name:
        get: return _name
    State:
        get: return _state
    Volume:
        get: return _volume
        set:
            _volume = Min(1, Max(0, value))
            lowest = 1.0f
            for i as SoundGroup in _groups:
                if i.Volume < lowest: lowest = i.Volume
            fslSoundSetGain(_soundObj, _volume * lowest)
    Pitch:
        get: return _pitch
        set:
            _pitch = Min(4, Max(0, value))
            if _groups.Count > 0:
                lowest = 4.0f
                for i as SoundGroup in _groups:
                    if i.Pitch < lowest: lowest = i.Pitch
            else: lowest = 1.0f
            fslSoundSetPitch(_soundObj,
                             _pitch * lowest * Audio.MasterPitch)
    SoundID:
        get: return _soundObj
    IsStreaming:
        get: return _stream
    IsLooping:
        get: return fslSoundIsLooping(_soundObj)
        set: fslSoundSetLooping(_soundObj, value)
    IsPlaying:
        get: return _state == SoundState.Playing
    IsPaused:
        get: return _state == SoundState.Paused
    IsStopped:
        get: return _state == SoundState.Stopped
    IsFading:
        get: return _fader.IsFading
    IsMuted:
        get: return _muted
    Groups as (SoundGroup):
        get: return _groups.ToArray(SoundGroup)
    PlayPosition:
        get: return fslSoundGetBufferPosition(_soundObj) / 1000.0f
        set:
            Pause()
            fslSoundSetBufferPosition(_soundObj, value * 1000)
            Unpause()
    PlayLength:
        get: return fslSoundGetBufferLength(_soundObj) / 1000.0f
    
    event Finished as callable(Sound)
    
    def constructor(name as string, soundObj as uint, stream as bool):
            
        _name = name
        _fader = VolumeFader(self)
        _state = SoundState.Stopped
        _stream = stream
        _soundObj = soundObj
    
    virtual def _update():
        
        if not fslSoundIsPlaying(_soundObj) and\
           not fslSoundIsPaused(_soundObj) and\
           _state == SoundState.Playing:
               Stop()
               Finished(self)
        
        if _fader.IsFading and _state != SoundState.Paused:
            _fader._update()
            
    def _joinGroup(group as SoundGroup):
        
        _groups.Add(group)
        
    def _leaveGroup(group as SoundGroup):
        
        _groups.Remove(group)
        
    def _forcePlayState():
        
        _state = SoundState.Playing
        
    def FadeIn(duration as single):
        
        FadeIn(duration, null)
        
    def FadeIn(duration as single, callback as callable(Sound)):
        
        Fade(0, 1, duration, callback)
        
    def FadeOut(duration as single):
        
        FadeOut(duration, null)
        
    def FadeOut(duration as single, callback as callable(Sound)):
        
        Fade(1, 0, duration, callback)
    
    def Fade(v1 as single, v2 as single, duration as single):
        
        Fade(v1, v2, duration, null)
    
    def Fade(v1 as single, v2 as single, duration as single,
             callback as callable(Sound)):
        
        if _state == SoundState.Paused: return
        if _fader.IsFading: _fader._cancelFading()
        _fader._fade(v1, v2, duration, callback)
        
    def CancelFading():
        
        _fader._cancelFading()
        
    def Play():
        
        Play(false)
        
    def Play(looping as bool):
        
        if _state == SoundState.Playing: Stop()
        elif _state == SoundState.Paused: 
            Unpause()
            return
        IsLooping = looping
        Audio._nominateSound(self)
        
    def PlayAsync():
        
        sound as Sound
        if self isa Sound3D:
            sound3d = sound as Sound3D
            self3d = self as Sound3D
            sound3d = Audio.Create3DSound(_name, self3d.Node)
            sound3d.IgnoreEnvironment = self3d.IgnoreEnvironment
            sound3d.InnerConeAngle = self3d.InnerConeAngle
            sound3d.OuterConeAngle = self3d.OuterConeAngle
            sound3d.OuterConeVolume = self3d.OuterConeVolume
            sound3d.MaxDistance = self3d.MaxDistance
            sound3d.MinDistance = self3d.MinDistance
            sound3d.Rolloff = self3d.Rolloff
            sound = sound3d
        else: sound = Audio.Create2DSound(_name)
        
        sound.Volume = _volume
        sound.Pitch = _pitch
        sound.Importance = _importance
        for i in Groups: i.AddSound(sound)

        _async.Add(sound)
        sound.Play()
        sound.Finished += {Audio._queueForDestruction(sound);
                           _async.Remove(sound)}
        
    def Stop():
        
        if _state == SoundState.Stopped: return
        fslSoundStop(_soundObj)
        fslSoundRewind(_soundObj)
        _state = SoundState.Stopped
        Audio._removeFromPlaylist(self)
        _fader._cancelFading()
        
    def Rewind():
        
        fslSoundRewind(_soundObj)
        
    def Pause():
        
        if _state != SoundState.Playing: return
        fslSoundPause(_soundObj)
        _state = SoundState.Paused
        
    def Unpause():
        
        if _state != SoundState.Paused: return
        fslSoundUnPause(_soundObj)
        _state = SoundState.Playing
        
    def Mute():
        
        fslSoundSetGain(_soundObj, 0)
        _muted = true
        
    def Unmute():
        
        v = Volume
        Volume = v
        _muted = false
        
class Sound3D(Sound):
    
    _prevPos as Vector3
    _ignoreEnv = false
    
    [Property(Node)] _source as IRenderable
    InnerConeAngle:
        get: return fslSoundGetConeInsideAngle(_soundObj)
        set: fslSoundSetConeInsideAngle(_soundObj, value)
    OuterConeAngle:
        get: return fslSoundGetConeOutsideAngle(_soundObj)
        set: fslSoundSetConeOutsideAngle(_soundObj, value)
    OuterConeVolume:
        get: return fslSoundGetConeOutsideVolume(_soundObj)
        set: fslSoundSetConeOutsideVolume(_soundObj, value)
    Rolloff:
        get: return fslSoundGetRolloff(_soundObj)
        set: fslSoundSetRolloff(_soundObj, value)
    MaxDistance:
        get: return fslSoundGetMaxDistance(_soundObj)
        set: fslSoundSetMaxDistance(_soundObj, value)
    MinDistance:
        get: return fslSoundGetReferenceDistance(_soundObj)
        set: fslSoundSetReferenceDistance(_soundObj, value)
    IgnoreEnvironment:
        get: return _ignoreEnv
        set:
            if not _ignoreEnv and value:
                env = fslGetCurrentBufferEnvironment(_soundObj)
                env.lDirectHF = -10000
                easySetBufferEnv(env, _soundObj)
            elif _ignoreEnv and not value:
                fslSetBufferDefaultEnvironment(_soundObj)
            _ignoreEnv = value
        
    def constructor(name as string, soundObj as uint,
                    node as IRenderable, streaming as bool):
        
        super(name, soundObj, streaming)
        
        _source = node
        if _source == null: _source = Graphics.SceneMgr.RootSceneNode
        _prevPos = _source.WorldPosition
        fslSoundSetConeOutsideVolume(_soundObj, 0)
        
    override def _update():
        
        super._update()
        
        if _source:
            
            pos = _source.WorldPosition
            vel = (pos - _prevPos) / Core.FrameTime
            ori = -_source.WorldOrientation.ZAxis
            _prevPos = pos
            pos *= Audio.DistanceScale
            
            fslSoundSetPosition(_soundObj, pos.x, pos.y, pos.z)
            fslSoundSetConeOrientation(_soundObj, ori.x, ori.y, ori.z)
            
            if Audio.IsDopplerEnabled:
            
                vel2 = vel.SquaredLength
                
                if vel2 >= Audio.MaxDopplerSpeed**2 or\
                   vel2 <= Audio.MinDopplerSpeed**2:
                    fslSoundSetVelocity(_soundObj, 0, 0, 0)
                else: fslSoundSetVelocity(_soundObj, vel.x, vel.y, vel.z)
    
class Sound2D(Sound):
    
    def constructor(name as string, soundObj as uint, streaming as bool):
        
        super(name, soundObj, streaming)
        
        fslSoundSetSourceRelative(_soundObj, true)
        env = fslGetCurrentBufferEnvironment(_soundObj)
        env.lDirectHF = -10000
        easySetBufferEnv(env, _soundObj)
        
class SoundGroup:
    
    _sounds = []
    _playlist = []
    _volume = 1.0f
    _pitch = 1.0f
    _fader as VolumeFader
    _paused = false
    _pausedSounds = []
    _muted = false
    _mutedSounds = []
    
    [Property(Name)] _name as string
    [Property(PlayLimit)] _playLimit as int
    [Property(LimitMode)] _limitMode as PlayLimitMode
    Volume:
        get: return _volume
        set:
            _volume = Max(0, Min(1, value))
            for i as Sound in _sounds:
            	v = i.Volume
            	i.Volume = v
    Pitch:
        get: return _pitch
        set:
            _pitch = Max(0, Min(4, value))
            for i as Sound in _sounds:
            	p = i.Pitch
            	i.Pitch = p
    Sounds as (Sound):
        get: return _sounds.ToArray(Sound)
    Playlist as (Sound):
        get: return _playlist.ToArray(Sound)
    IsFading:
        get: return _fader.IsFading
    IsPaused:
        get: return _paused
    IsMuted:
        get: return _muted
    Importance:
        set:
            for i as Sound in _sounds: i.Importance = value
        
    def constructor(name as string, playLimit as int,
                    limitMode as PlayLimitMode):
        
        _name = name
        _playLimit = playLimit
        _limitMode = limitMode
        _fader = VolumeFader(self)
                
    def _update():
        
        if not _paused and _fader.IsFading: _fader._update()
                
    def _addToPlaylist(sound as Sound):

        _playlist.Add(sound)
                
    def _removeFromPlaylist(sound as Sound):
        
        _playlist.Remove(sound)
        
    def AddSound(sound as Sound):
        
        AddSound(sound, _sounds.Count)
        
    def AddSound(sound as Sound, index as int):
        
        _sounds.Insert(index, sound)
        sound._joinGroup(self)
        
        v = sound.Volume
        sound.Volume = v
        
        if _paused:
            if sound.State != SoundState.Paused:
                sound.Pause()
                _pausedSounds.Add(sound)
        if _muted:
            if not sound.IsMuted:
                sound.Mute()
                _mutedSounds.Add(sound)

    def ReplaceSound(sound as Sound, index as int):
        
        RemoveSound(index)
        AddSound(sound, index)
        
    def ReplaceAllSounds(sounds as (Sound)):
        
        ClearSounds()
        for i in sounds: AddSound(i)
        
    def RemoveSound(sound as Sound):
        
        _sounds.Remove(sound)
        sound._leaveGroup(self)
        
        v = sound.Volume
        sound.Volume = v
        
        _playlist.Remove(sound)
        _pausedSounds.Remove(sound)
        _mutedSounds.Remove(sound)
        
    def RemoveSound(index as int):
        
        RemoveSound(_sounds[index] as Sound)
        
    def ClearSounds():
        
        while _sounds.Count > 0: RemoveSound(0)
        
    def FadeIn(duration as single):
        
        FadeIn(duration, null)
        
    def FadeIn(duration as single, callback as callable(Sound)):
        
        Fade(0, 1, duration, callback)
        
    def FadeOut(duration as single):
        
        FadeOut(duration, null)
        
    def FadeOut(duration as single, callback as callable(Sound)):
        
        Fade(1, 0, duration, callback)
    
    def Fade(v1 as single, v2 as single, duration as single):
        
        Fade(v1, v2, duration, null)
    
    def Fade(v1 as single, v2 as single, duration as single,
             callback as callable(Sound)):
        
        if _fader.IsFading: _fader._cancelFading()
        _fader._fade(v1, v2, duration, callback)
        
    def CancelFading():
        
        _fader._cancelFading()
        
    def Play():
        
        Play(false)
        
    def Play(looping as bool):
        
        if _paused:
            Unpause()
            return
        for i as Sound in _sounds: i.Play(looping)
        
    def Stop():
        
        for i as Sound in _sounds: i.Stop()
        
    def Rewind():
        
        for i as Sound in _sounds: i.Rewind()
        
    def Pause():
        
        for i as Sound in _sounds:
            if not i.State == SoundState.Paused: 
                i.Pause()
                _pausedSounds.Add(i)
        _paused = true
                
    def Unpause():
        
        for i as Sound in _pausedSounds: i.Unpause()
        _pausedSounds.Clear()
        _paused = false
        
    def Mute():
        
        for i as Sound in _sounds:
            if not i.IsMuted: 
                i.Mute()
                _mutedSounds.Add(i)
        _muted = true
        
    def Unmute():
        
        for i as Sound in _mutedSounds: i.Unmute()
        _mutedSounds.Clear()
        _muted = false

[Module]
static class Audio:
    
    _defaults = {}
    _sounds = []
    _playlist = []
    _loadedSounds = {}
    _groups = []
    _pausedSounds = []
    _destroyQueue = []
    _paused = false
    _muted = false
    _masterVolume = 1.0f
    _masterPitch = 1.0f
    _musicGroup as SoundGroup
    _interfaceGroup as SoundGroup
    _effectsGroup as SoundGroup
    _distanceModel as SoundDistanceModel
    _updateThread as Thread
    _prevListenerPos as Vector3
    _environment = FSL_LISTENER_ENVIRONMENT.FSL_ENVIRONMENT_COUNT
    _envEnabled = true
    _dopEnabled = true

    [Property(PlayLimit)] _playLimit as uint
    [Property(DistanceScale)] _scale = 1.0f
    [Property(ListenerNode)] _node as IRenderable
    [Property(MaxDopplerSpeed)] _maxDopSpeed = 500.0f
    [Property(MinDopplerSpeed)] _minDopSpeed = 0.5f
    [Property(MaxLoadSize)] _maxLoadSize as uint = 200
    IsDopplerEnabled:
        get: return _dopEnabled
        set:
            for i as Sound in _sounds:
                if i isa Sound3D: fslSoundSetVelocity(i.SoundID, 0, 0, 0)
            _dopEnabled = value
    IsEnvRenderingEnabled:
        get: return _envEnabled
        set:
            if value:
            	e = Environment
            	Environment = e
            else: fslSetListenerDefaultEnvironment()
            _envEnabled = value
    CurrentConfig as Hash:
        get:
            volumes = {}
            volumes['master'] = _masterVolume
            volumes['music'] = _musicGroup.Volume
            volumes['interface'] = _interfaceGroup.Volume
            volumes['effects'] = _effectsGroup.Volume
            config = {}
            config['volume'] = volumes
            config['env_rendering'] = _envEnabled
            config['doppler'] = _dopEnabled
            return config
    DefaultConfig as Hash:
        get: return _defaults.Clone()
    Playlist as (Sound):
        get: return _playlist.ToArray(Sound)
    Sounds as (Sound):
        get: return _sounds.ToArray(Sound)
    MasterVolume:
        get: return _masterVolume
        set:
            _masterVolume = Max(0, Min(1, value))
            fslSetVolume(_masterVolume)
    MasterPitch:
        get: return _masterPitch
        set:
            _masterPitch = Max(0, Min(4, value))
            for i as Sound in _sounds:
            	p = i.Pitch
            	i.Pitch = p
    MusicGroup:
        get: return _musicGroup
    InterfaceGroup:
        get: return _interfaceGroup
    EffectsGroup:
        get: return _effectsGroup
    IsMuted:
        get: return _muted
    IsPaused:
        get: return _paused
    DistanceModel:
        get: return _distanceModel
        set:
            _distanceModel = value
            fslSetListenerDistanceModel(cast(uint, _distanceModel))
    Environment:
        get: return cast(SoundEnvironment, cast(int, _environment))
        set:
            _environment = cast(FSL_LISTENER_ENVIRONMENT, cast(int, value))
            if value == SoundEnvironment.None:
                fslSetListenerDefaultEnvironment()
            else:
                if _envEnabled:
                    fslSetListenerEnvironmentPreset(_environment)
       
    def _nominateSound(sound as Sound):
            
        canPlay = true
        
        for j in sound.Groups:
            if j.Playlist.Length == j.PlayLimit and j.PlayLimit != 0:
                if j.LimitMode == PlayLimitMode.RejectNew:
                    canPlay = false
                else: j.Playlist[0].Stop()
                
        if not canPlay: return
        
        lowerBound = upperBound = _playlist.Count
        
        if sound isa Sound3D:
            
            for j in range(_playlist.Count):
                if _playlist[j] isa Sound3D:
                    upperBound = j
                    break
                    
            s = sound as Sound3D
                
            offset = _node.WorldPosition - s.Node.WorldPosition
            dist2 = Max(0.01f, offset.SquaredLength)
            finalImp = s.Importance**2 / dist2
                
            for j in range(upperBound, lowerBound):
                
                c = _playlist[j] as Sound3D
                offset = _node.WorldPosition - c.Node.WorldPosition
                dist2 = Max(0.01f, offset.SquaredLength)
                competitorImp = c.Importance**2 / dist2
                
                if competitorImp < finalImp:
                    upperBound = j
                    break
        
        else:
            
            for j in range(_playlist.Count):
                if _playlist[j] isa Sound3D:
                    lowerBound = j
                    break
            
            upperBound = 0
                
            for j in range(upperBound, lowerBound):
                if (_playlist[j] as Sound2D).Importance <\
                   (sound as Sound2D).Importance:
                    upperBound = j
                    break
           
        if _playlist.Count == _playLimit: 
            if upperBound == _playlist.Count: return
            (_playlist.Pop() as Sound).Stop()
        
        fslSoundPlay(sound.SoundID)
        sound._forcePlayState()
        _playlist.Insert(upperBound, sound)
        for j in sound.Groups: j._addToPlaylist(sound)
        
    private def _asyncUpdate():
        
        while 1:
            try:
                fslUpdate()
                fslSleep(0.05f)
            except: break
            
    def _addToLoaded(name as string, fileName as string,
                     zipName as string, stream as bool, soundObj as uint):
        
        _loadedSounds.Add(name, (fileName, zipName, stream, soundObj))
        
    def _removeFromLoaded(soundObj as uint):
        
        for i as string in _loadedSounds.Keys:
            if (_loadedSounds[i] as (object))[3] == soundObj:
                _loadedSounds.Remove(i)
                return
        
    def _removeFromPlaylist(sound as Sound):
        
        _playlist.Remove(sound)
        for i as SoundGroup in sound.Groups: i._removeFromPlaylist(sound)
        
    def _queueForDestruction(sound as Sound):
        
        _destroyQueue.AddUnique(sound)
    
    def _setup(config as Hash, defaults as Hash):
        
        fslInit(FSL_SOUND_SYSTEM.FSL_SS_EAX2)
        _playLimit = fslGetMaxSources()
        
        _node = Graphics.MainCamera
        _prevListenerPos = _node.WorldPosition
        
        Graphics.Form.GotFocus += {x, y | Unmute()}
        Graphics.Form.LostFocus += {x, y | Mute()}
        Graphics.Form.SizeChanged += {x, y | (Unpause, Pause)\
        [Graphics.Form.WindowState == FormWindowState.Minimized]()}

        
        _musicGroup = SoundGroup('music', 2, PlayLimitMode.RejectNew)
        _effectsGroup = SoundGroup('effects', 0,PlayLimitMode.RejectNew)
        _interfaceGroup = SoundGroup('interface', 2,
                                     PlayLimitMode.ReplaceOld)
        
        _defaults = defaults
        ApplyConfig(config)
        
        Content.AddType('sound', ('*.ogg', '*.wav'), SoundLoader())
        
        _updateThread = Thread(ThreadStart(_asyncUpdate))
        _updateThread.Start()
        
    def _update():
        
        if _paused: return
        
        pos = _node.WorldPosition
        vel = (pos - _prevListenerPos) / Core.FrameTime
        ori = _node.WorldOrientation
        _prevListenerPos = pos
        
        pos *= _scale
        
        fslSetListenerPosition(pos.x, pos.y, pos.z)
        fslSetListenerOrientation(-ori.ZAxis.x, -ori.ZAxis.y,
                                  -ori.ZAxis.z, ori.YAxis.x,
                                  ori.YAxis.y, ori.YAxis.z)
 
        if _dopEnabled:
               
            vel2 = vel.SquaredLength
            if vel2 >= _maxDopSpeed**2 or vel2 <= _minDopSpeed**2:
                fslSetListenerVelocity(0, 0, 0)
            else: fslSetListenerVelocity(vel.x, vel.y, vel.z)
                                  
        for i as Sound in _sounds: i._update()
        for i as SoundGroup in _groups: i._update()
        
        _musicGroup._update()
        _effectsGroup._update()
        _interfaceGroup._update()
        
        for i as Sound in _destroyQueue: DestroySound(i)
        _destroyQueue.Clear()
        
    def _shutdown():
        
        _updateThread.Abort()
        ClearSoundGroups()
        ClearSounds()
        fslShutDown()
        
    def ApplyConfig(config as Hash):
        
        volumes = config['volume'] as Hash
        MasterVolume = volumes['master']
        _musicGroup.Volume = volumes['music']
        _interfaceGroup.Volume = volumes['interface']
        _effectsGroup.Volume = volumes['effects']
        _envEnabled = config['env_rendering']
        _dopEnabled = config['doppler']
        
    def Mute():
        
        if _muted: return
        fslSetVolume(0)
        _muted = true
        
    def Unmute():

        if not _muted: return
        fslSetVolume(_masterVolume)
        _muted = false
        
    def StopAll():
        
        for i as Sound in _playlist: i.Stop()
        
    def Pause():

        if _paused: return
        for i as Sound in _playlist:
            if i.State != SoundState.Playing: continue
            i.Pause()
            _pausedSounds.Add(i)
        _paused = true
        
    def Unpause():
    
        if not _paused: return
        for i as Sound in _pausedSounds: i.Unpause()
        _pausedSounds.Clear()
        _paused = false
        
    def IsSoundLoaded(name as string):
        
        return _loadedSounds.ContainsKey(name)
        
    def Create3DSound(name as string):
        
        return Create3DSound(name, null)
        
    def Create3DSound(name as string, node as IRenderable):
        
        loadedEntry = _loadedSounds[name.ToLower()] as (object)
        fileName as string = loadedEntry[0]
        zipName as string = loadedEntry[1]
        streaming as bool = loadedEntry[2]
        
        if zipName == '':
            if streaming: soundObj = fslStreamSound(fileName)
            else: soundObj = fslLoadSound(fileName)
        else: soundObj = fslLoadSoundFromZip(zipName, fileName)

        sound = Sound3D(name, soundObj, node, streaming)
        _sounds.Add(sound)
        return sound
        
    def Create2DSound(name as string):
        
        loadedEntry = _loadedSounds[name.ToLower()] as (object)
        fileName as string = loadedEntry[0]
        zipName as string = loadedEntry[1]
        streaming as bool = loadedEntry[2]

        if zipName == '':
            if streaming: soundObj = fslStreamSound(fileName)
            else: soundObj = fslLoadSound(fileName)
        else: soundObj = fslLoadSoundFromZip(zipName, fileName)

        sound = Sound2D(name, soundObj, streaming)
        _sounds.Add(sound)
        return sound
        
    def CreateSoundGroup(name as string):
        
        return CreateSoundGroup(name, 0)
        
    def CreateSoundGroup(name as string, playLimit as int):
        
        return CreateSoundGroup(name, playLimit, PlayLimitMode.RejectNew)
        
    def CreateSoundGroup(name as string, playLimit as int,
                         limitMode as PlayLimitMode):
        
        for i as SoundGroup in _groups:
            if i.Name == name: return null
        group = SoundGroup(name, playLimit, limitMode)
        _groups.Add(group)
        return group
        
    def GetSoundGroup(name):
        
        for i as SoundGroup in _groups:
            if i.Name == name: return i
        return null
        
    def DestroySound(sound as Sound):
        
        sound.Stop()
        for i in sound.Groups: i.RemoveSound(sound)
        if not sound.IsStreaming: fslFreeSound(sound.SoundID, true)
        _sounds.Remove(sound)
    
    def ClearSounds():
        
        for i as Sound in _sounds:
            i.Stop()
            for j in i.Groups: j.RemoveSound(i)
            if not i.IsStreaming: fslFreeSound(i.SoundID, true)
        _sounds.Clear()
        
    def DestroySoundGroup(group as SoundGroup):
        
        _groups.Remove(group)
        
    def DestroySoundGroup(name as string):
        
        for i as SoundGroup in _groups:
            if i.Name == name:
                _groups.Remove(i)
                return
    
    def ClearSoundGroups():

        _groups.Clear()
        