//=================================================================================================
//=================== class implementations: fpModel and fpHierarchyAllocator =====================
//=================================================================================================
//
//      classes for loading .x files for use as static or animated models. supports skinned
//      meshes and multiple animations as well as animation blending.
//
//=================================================================================================


#include "fpModel.h"


LPDIRECT3DDEVICE9 device;    // global device


///////////////////////////////////////////////////////////////////////////////////////////////////

void AllocateName( LPCSTR in_name, LPSTR* out_name )                           // allocates names
{
    UINT length;

    if( in_name != NULL )                                                      // change name to match in_name
    {
        length = ( UINT )strlen( in_name ) + 1;
        *out_name = new char[ length ];
        memcpy( *out_name, in_name, length * sizeof( char ) );
    }
    else *out_name = NULL;                                                     // no name specified
}

///////////////////////////////////////////////////////////////////////////////////////////////////

HRESULT fpHierarchyAllocator::CreateFrame( LPCSTR name, LPD3DXFRAME* new_frame )      // create new frame
{
    fpFrame* frame = new fpFrame;                                // temporary frame
    AllocateName( name, &frame->Name );                          // allocate a name

    D3DXMatrixIdentity( &frame->TransformationMatrix );          // reset the members
    D3DXMatrixIdentity( &frame->combined_matrix );               // ~

    frame->pMeshContainer = NULL;                                // ~
    frame->pFrameSibling = NULL;                                 // ~
    frame->pFrameFirstChild = NULL;                              // ~

    *new_frame = frame;                                          // point new frame to temporary frame
    frame = NULL;                                                // unpoint temporary frame

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

HRESULT fpHierarchyAllocator::CreateMeshContainer( LPCSTR name,                               // create a new mesh container
                                                   const D3DXMESHDATA* mesh_data,
												   const D3DXMATERIAL* materials, 
												   const D3DXEFFECTINSTANCE* effect, 
												   DWORD num_materials, 
												   const DWORD* adjacency, 
												   LPD3DXSKININFO skin_info, 
												   LPD3DXMESHCONTAINER* new_mesh_container ) 
{
    fpMeshContainer* mesh_container = new fpMeshContainer;                   // temporary mesh container                                                  
    LPD3DXMESH mesh = mesh_data->pMesh;                                      // get the mesh from mesh_data

    memset( mesh_container, 0, sizeof( fpMeshContainer ) );                  // zero out temporary mesh container

    AllocateName( name, &mesh_container->Name );                             // allocate name for mesh container      

    UINT num_faces = mesh->GetNumFaces();                                    // retrieve number of faces

    mesh_container->NumMaterials = max( 1, num_materials );                  // if no materials, create one
    mesh_container->pAdjacency   = new DWORD[ num_faces * 3 ];               // create adjacency array with 3 elements for each face

    memcpy( mesh_container->pAdjacency, adjacency, sizeof( DWORD ) * num_faces * 3 );                        // copy over adjacency data

    if( !( mesh->GetFVF() & D3DFVF_NORMAL ) )                                // if mesh does not have normals...
    {
        mesh->CloneMeshFVF( mesh->GetOptions(),                              // add normals to the vertex format
                            mesh->GetFVF() | 
							D3DFVF_NORMAL, 
                            device, 
							&mesh );

        D3DXComputeNormals( mesh, mesh_container->pAdjacency );              // generate normals for the mesh
    }
	else mesh->AddRef();                                                     

    mesh_container->MeshData.pMesh = mesh;                                   // assign mesh to the temporary mesh container's mesh 
    mesh_container->MeshData.Type = D3DXMESHTYPE_MESH;
        
    if( skin_info )															 // if there's skinning info
    {
        mesh_container->pSkinInfo = skin_info;								 // save off the skin info
        skin_info->AddRef();

        mesh_container->original_mesh = mesh;								 // I really can't figure out what this is for.
        mesh->AddRef();

        UINT num_bones = skin_info->GetNumBones();                           // retrieve number of bones
        mesh_container->offset_matrices = new D3DXMATRIX[ num_bones ];       // create offset matrix array with num_bones

        for( UINT i = 0; i < num_bones; i++ )                                // save all the offset matrices
        {
            mesh_container->offset_matrices[ i ] = *( mesh_container->pSkinInfo->GetBoneOffsetMatrix( i ) );
        }

		D3DCAPS9 caps;
		device->GetDeviceCaps( &caps );                                      // get device capabilities

		if( mesh_container->MeshData.pMesh ) mesh_container->MeshData.pMesh->Release();             
		if( mesh_container->bone_combo_buffer ) mesh_container->bone_combo_buffer->Release();

		DWORD max_influences;                                                // maximum number of face influences                

		LPDIRECT3DINDEXBUFFER9 ib;                                           // get index buffer                        
		mesh_container->original_mesh->GetIndexBuffer( &ib );

		mesh_container->pSkinInfo->GetMaxFaceInfluences( ib, mesh_container->original_mesh->GetNumFaces(), &max_influences );  // retrieve maximum number of influences
		ib->Release();

		max_influences = min( max_influences, 12 );                          // if the max is too great, use 12, 4 for each vertex of a triangle.

		mesh_container->num_palette_entries = min( ( caps.MaxVertexBlendMatrixIndex + 1 ) / 2,     // if the number of matrix palette entries is too great, use the device maximum
												mesh_container->pSkinInfo->GetNumBones() );

		mesh_container->pSkinInfo->ConvertToIndexedBlendedMesh( mesh_container->original_mesh,     // conver to indexed, skinned mesh
																D3DXMESHOPT_VERTEXCACHE |  
																D3DXMESH_SYSTEMMEM, 
																mesh_container->num_palette_entries, 
																mesh_container->pAdjacency, 
																NULL, 
																NULL, 
																NULL, 
																&mesh_container->num_influences,
																&mesh_container->num_att_groups, 
																&mesh_container->bone_combo_buffer, 
																&mesh_container->MeshData.pMesh );
    }

    *new_mesh_container = mesh_container;									 // pass out the temporary mesh container as the new mesh container
	mesh_container = NULL;													 // unpoint the temporary mesh container

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

HRESULT fpHierarchyAllocator::DestroyFrame( LPD3DXFRAME frame )    // destroy a frame
{
    if( frame->Name ) delete[] frame->Name; 
    if( frame ) delete frame;      

    return S_OK; 
}

///////////////////////////////////////////////////////////////////////////////////////////////////

HRESULT fpHierarchyAllocator::DestroyMeshContainer( LPD3DXMESHCONTAINER mesh_container )       // destroy a mesh container
{
    fpMeshContainer* temp_mc = ( fpMeshContainer* )mesh_container;

	if( temp_mc->Name ) delete[] temp_mc->Name;                         // release other crap
    if( temp_mc->pAdjacency ) delete[] temp_mc->pAdjacency;
    if( temp_mc->pMaterials ) delete[] temp_mc->pMaterials;
    if( temp_mc->offset_matrices ) delete[] temp_mc->offset_matrices;  
    if( temp_mc->combined_matrices ) delete[] temp_mc->combined_matrices;
    if( temp_mc->bone_combo_buffer ) temp_mc->bone_combo_buffer->Release();
    if( temp_mc->MeshData.pMesh ) temp_mc->MeshData.pMesh->Release();
    if( temp_mc->pSkinInfo ) temp_mc->pSkinInfo->Release();
    if( temp_mc->original_mesh ) temp_mc->original_mesh->Release();
    if( temp_mc ) delete temp_mc;

    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

bool fpModel::Create( LPDIRECT3DDEVICE9 temp_device,             // create the model
					  const char* model_file )      
{
	halloc = new fpHierarchyAllocator;                           // hierarchy allocator

	device = temp_device;
 
	if( FAILED( D3DXLoadMeshHierarchyFromX( model_file,          // attempt to load mesh
		                                    D3DXMESH_SYSTEMMEM, 
											device, 
											halloc, 
											NULL, 
											&root, 
											&animator ) ) )
	{
		Release();
		return FALSE;
	}

	SetupMatrices( root, root );                                             // setup matrices

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void fpModel::Release()                         // release the model
{
    D3DXFrameDestroy( root, halloc ); 
    if( animator ) animator->Release();
	if( halloc ) delete halloc;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void SetupMatrices( LPD3DXFRAME root, LPD3DXFRAME frame )              // setup combined matrices
{          
	if( !frame || !root ) return;

	if( frame->pMeshContainer )                                        // if there is a mesh container...
	{
		fpMeshContainer* cmesh = ( fpMeshContainer* )frame->pMeshContainer;  

		if( cmesh->pSkinInfo )                                         // if there is skinning info...
		{ 
			UINT num_bones = cmesh->pSkinInfo->GetNumBones();          // retrieve number of bones that affect the mesh

			cmesh->combined_matrices = new D3DXMATRIX*[ num_bones ];   
	 
			for( UINT i = 0; i < num_bones; i ++ )                     // loop through all the bones
			{
				fpFrame* cframe = ( fpFrame* )D3DXFrameFind( root, cmesh->pSkinInfo->GetBoneName( i ) );

				cmesh->combined_matrices[ i ] = &cframe->combined_matrix;              // get combined matrix, and save a pointer
			}
		}
	}

	if( frame->pFrameSibling ) SetupMatrices( root, frame->pFrameSibling );             
	if( frame->pFrameFirstChild ) SetupMatrices( root, frame->pFrameFirstChild );
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void UpdateMatrices( LPD3DXFRAME frame, LPD3DXMATRIX parent )        // update combined matrices
{
	fpFrame* cframe = ( fpFrame* )frame;   

	D3DXMatrixMultiply( &cframe->combined_matrix, &cframe->TransformationMatrix, parent );         // multiply local by parent's combined

	if( cframe->pFrameSibling ) UpdateMatrices( cframe->pFrameSibling, parent );
	if( cframe->pFrameFirstChild ) UpdateMatrices( cframe->pFrameFirstChild, &cframe->combined_matrix );
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void fpModel::Render(  LPDIRECT3DDEVICE9 device )              // render the model
{
	D3DXMATRIX matrix;
	device->GetTransform( D3DTS_WORLD, &matrix );
	UpdateMatrices( root, &matrix );                           // update matrices using current world transform matrix

	DrawMesh( device, root );                                  // draw the mesh
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void DrawMesh(  LPDIRECT3DDEVICE9 device, LPD3DXFRAME frame )                         // draw mesh
{
	fpMeshContainer* cmesh = ( fpMeshContainer* )frame->pMeshContainer;               // make derived version of container

	while( cmesh )                                                                    // if there is a mesh container...
	{
		fpFrame* cframe = ( fpFrame* )frame;                                          // make derived version of frame

		if( frame->pMeshContainer->pSkinInfo )                                        // if there is skinning info...
		{
            device->SetRenderState( D3DRS_VERTEXBLEND, cmesh->num_influences - 1 );   // set vertex blending stuff
            device->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, TRUE );           // enable indexed vertex blending

			D3DXMATRIX temp_matrix;                                                   // temporary matrix

            LPD3DXBONECOMBINATION bone_combo_buffer = ( LPD3DXBONECOMBINATION )cmesh->bone_combo_buffer->GetBufferPointer();  // get combination buffer

            for( UINT i = 0; i < cmesh->num_att_groups; i ++ )                        // loop through all the subsets
            {
                for( UINT j = 0; j < cmesh->num_palette_entries; j ++ )               // loop through every bone in the subset
                {
                    int matrix_id = bone_combo_buffer[ i ].BoneId[ j ];               // retrieve matrix for that bone

                    D3DXMatrixMultiply( &temp_matrix,                                 // apply offset
                                        &cmesh->offset_matrices[ matrix_id ], 
                                        cmesh->combined_matrices[ matrix_id ] );

                    device->SetTransform( D3DTS_WORLDMATRIX( j ), &temp_matrix );     // transform 
                }
                    
                cmesh->MeshData.pMesh->DrawSubset( i );                                                     // draw skinned mesh
            }

            device->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE );                                // reset settings
            device->SetRenderState( D3DRS_VERTEXBLEND, D3DVBF_DISABLE );
		}
		else
		{
			device->SetTransform( D3DTS_WORLD, &cframe->combined_matrix );            // no skinning. draw static mesh.

			for( UINT i = 0; i < cmesh->NumMaterials; i ++ )                 
			{
				cmesh->MeshData.pMesh->DrawSubset(i);
			}
		}

		cmesh = ( fpMeshContainer* )frame->pMeshContainer->pNextMeshContainer;        // next mesh if required
	}

	if( frame->pFrameFirstChild ) DrawMesh( device, frame->pFrameFirstChild );        // recurse down the hierarchy 
	if( frame->pFrameSibling ) DrawMesh( device, frame->pFrameSibling );              // recurse sideways through the hierarchy
}

