#include <omgUtility.h>
#include "panGame.h"
#include "panPlayer.h"

using namespace Ogre;
using namespace omg;
using namespace std;

namespace pan
{
	Player::Player()
	{
		addType( "player" );

		bodyGeo = 0;
		shellGeo = 0;

		speed = 6;
		mass = 85;
		height = 1.8;
		radius = 0.6;
		impulse = 0.001;
		jumpAmount = 700;

		crouching = false;
		jumpingUp = false;

		zoomed = false;

		hitByLow = false;
		hitByMedium = false;
		hitByHigh = false;

		absolutePitch = 0;

		deadTime = 0;
		fallTime = 0;

		// stats
		blood = 1;
		cobalt = 1;

		// weapons
		currentWeapon = 0;
		hydraRifle = 0;
	}

	Player::~Player()
	{
	}

	void Player::collision( mec::InformativeContact* inContact )
	{
		if( inContact->t2 == GT_CONCRETE ||
			inContact->t2 == GT_PLASTIC ||
			inContact->t2 == GT_GLASS ||
			inContact->t2 == GT_HOLOGRAM ||
			inContact->t2 == GT_DIRT ||
			inContact->t2 == GT_METAL )
		{
			if( inContact->g1 == shellGeo )
			{
				environmentNormal += inContact->normal;

				if( inContact->normal.y > 0.75 ) canReachGround = true;
			}
			else
			{
				if( inContact->normal.y > 0.75 ) touchingGround = true;
			}
		}
		else if( inContact->t2 == GT_ENEMY_FIRE_LOW && !hitByLow )
		{
			blood -= 0.01;
			hitByLow = true;
			if( !fslSoundIsPlaying( gGame->guiSoundSource ) ) fslSoundPlay( damageSound, gGame->guiSoundSource );
		}
	}

	void Player::reference()
	{
		sm = mec::gGraphics->getSceneManager();
		cam = sm->getCamera( "camera" );

		moveLeft = mec::gInput->getActionSignalGroup( "moveLeft" );
		moveRight = mec::gInput->getActionSignalGroup( "moveRight" );
		moveForward = mec::gInput->getActionSignalGroup( "moveForward" );
		moveBackward = mec::gInput->getActionSignalGroup( "moveBackward" );
		lookLeft = mec::gInput->getActionSignalGroup( "lookLeft" );
		lookRight = mec::gInput->getActionSignalGroup( "lookRight" );
		lookUp = mec::gInput->getActionSignalGroup( "lookUp" );
		lookDown = mec::gInput->getActionSignalGroup( "lookDown" );
		jump = mec::gInput->getActionSignalGroup( "jump" );
		crouch = mec::gInput->getActionSignalGroup( "crouch" );
		fire = mec::gInput->getActionSignalGroup( "fire" );
		zoom = mec::gInput->getActionSignalGroup( "zoom" );

		death = ( ( MaterialPtr )MaterialManager::getSingleton().getByName( "death" ) )->
			    getTechnique( 0 )->getPass( 0 )->getTextureUnitState( 0 );
		death->setTextureUScroll( -0.5 );

		bloodMat = ( ( MaterialPtr )MaterialManager::getSingleton().getByName( "bloodGauge" ) )->
			       getTechnique( 0 )->getPass( 0 )->getTextureUnitState( 0 );

		cobaltMat = ( ( MaterialPtr )MaterialManager::getSingleton().getByName( "cobaltGauge" ) )->
			        getTechnique( 0 )->getPass( 0 )->getTextureUnitState( 0 );
	}

	bool Player::setup( mec::Serial* inSerial )
	{
		body = new OgreOde::Body;
		body->setAutoSleep( false );

		bodyNode = sm->getRootSceneNode()->createChildSceneNode( "playerBodyNode" );
		bodyNode->attachObject( body );

		eyeNode = bodyNode->createChildSceneNode( "playerEyeNode" );
		eyeNode->attachObject( cam );
		eyeNode->setPosition( 0, 0, - height / 2 + 0.2 );
		eyeNode->pitch( Degree( -90 ) );

		setupGeometry();
/*
		groundSeeker = new OgreOde::RayGeometry( 500, gPhysics->getWorld()->getDefaultSpace() );
		groundSeeker->setBody( body );
		groundSeeker->setMaxContacts( 1 );
		addGeometry( groundSeeker, GT_TESTER );
*/
		body->setPosition( StringConverter::parseVector3( inSerial->getAttribute( "position" ) ) );
		desiredOrientation = StringConverter::parseQuaternion( inSerial->getAttribute( "orientation" ) ) * 
			                 Quaternion( Degree( 90 ), Vector3::UNIT_X ) * 
							 Quaternion( Degree( 180 ), Vector3::UNIT_Z );
		body->setOrientation( desiredOrientation );

		OverlayManager::getSingleton().getByName( "hud" )->show();
		OverlayManager::getSingleton().getByName( "hudContents" )->show();
		Overlay* ov = OverlayManager::getSingleton().getByName( "deathPanel" );
		ov->show();
		ov->getChild( "death" )->getChild( "deathText" )->setCaption( "" );

		mec::Serial* weapons = inSerial->getSubSerial( "weapons" );
		if( parseBoolean( weapons->getAttribute( "hydraRifle" ) ) ) 
		{
			hydraRifle = new HydraRifle( this );
			currentWeapon = hydraRifle;
			if( hydraRifle && !hydraRifle->setup() ) return false;
			currentWeapon->use();
		}

		zoomSound = mec::gAudio->getSound( "zoom.wav" );
		unzoomSound = mec::gAudio->getSound( "unzoom.wav" );
		damageSound = mec::gAudio->getSound( "damage.wav" );

		return true;
	}

	void Player::setupGeometry( bool inCrouching )
	{
		if( bodyGeo )
		{
			removeGeometry( bodyGeo );
			delete bodyGeo;
		}
		bodyGeo = new OgreOde::CapsuleGeometry( radius, 
												max( 0.01, height / ( inCrouching ? 2 : 1 ) - radius * 2 ), 
												mec::gPhysics->getWorld()->getDefaultSpace() );
		bodyGeo->setBody( body );
		addGeometry( bodyGeo, GT_PLAYER );

		body->setMass( OgreOde::CapsuleMass( mass, 
			                                 radius, 
											 Vector3::UNIT_Y, 
											 max( 0.01, height / ( inCrouching ? 2 : 1 ) - radius * 2 ) ) );

		if( shellGeo )
		{
			removeGeometry( shellGeo );
			delete shellGeo;
		}
		shellGeo = new OgreOde::CapsuleGeometry( radius + 0.2, 
												 max( 0.01, height / ( inCrouching ? 2 : 1 ) - radius * 2 ), 
												 mec::gPhysics->getWorld()->getDefaultSpace() );
		shellGeo->setBody( body );
		addGeometry( shellGeo, GT_TESTER );
	}

	void Player::shutdown()
	{
		if( hydraRifle )
		{
			hydraRifle->shutdown();
		    delete hydraRifle;
		}

		removeAllGeometry();

		OverlayManager::getSingleton().getByName( "hud" )->hide();
		OverlayManager::getSingleton().getByName( "hudContents" )->hide();
		OverlayManager::getSingleton().getByName( "deathPanel" )->hide();

		//delete groundSeeker;
		delete bodyGeo;
		delete shellGeo;
		delete body;

		sm->destroySceneNode( "playerEyeNode" );
		sm->destroySceneNode( "playerBodyNode" );
	}

	void Player::update()
	{
		static float stableFactor = 1;
	
		if( mec::gInput->getKeyboard()->getKey( mec::KC_F )->getCeased() ) speed = speed == 25 ? 6 : 25;

		fslSetListenerPosition( cam->getWorldPosition().x, 
			                    cam->getWorldPosition().y, 
								cam->getWorldPosition().z );
		fslSetListenerOrientation( cam->getDirection().x, 
			                       cam->getDirection().y, 
								   cam->getDirection().z,
			                       cam->getUp().x,
			                       cam->getUp().y,
			                       cam->getUp().z );

		bloodMat->setTextureUScroll( ( 1 - blood ) / 2 );
		cobaltMat->setTextureUScroll( ( 1 - cobalt ) / 2 );

		if( blood <= 0 )
		{
			sm->setFog( FOG_LINEAR, ColourValue( 0.3, 0, 0 ), 0, 0, 300 );
			sm->setSkyBox( false, "" );

			body->setDamping( 500, 300 );
			
			if( !deadTime )
			{
				OverlayManager::getSingleton().getByName( "deathPanel" )->getChild( "death" )->
				getChild( "deathText" )->setCaption( "FATALITY" );
				fslSoundPlay( mec::gAudio->getSound( "dead.wav" ), gGame->musicSoundSource );
			}

			fslSoundSetGain( gGame->musicSoundSource, sin( deadTime / 2 ) );

			deadTime += mec::gTime->getDelta();

			death->setTextureUScroll( min( 0.5, ( deadTime - 2 ) / 5 - 0.5 ) );
			if( deadTime >= 3 ) gGame->fatality = true;
			if( deadTime >= 6.28 ) fslSoundStop( gGame->musicSoundSource );

			return;
		}

		if( zoom->getCeased() ) 
		{
			zoomed = !zoomed;
			if( zoomed ) fslSoundPlay( zoomSound, gGame->guiSoundSource );
			else fslSoundPlay( unzoomSound, gGame->guiSoundSource );
		}

		if( zoomed ) 
		{
			stableFactor = 0.4;
			if( cam->getFOVy() > Ogre::Degree( 35 ) ) cam->setFOVy( cam->getFOVy() - Ogre::Degree( 5 ) );
		}
		else
		{
			stableFactor = 1;
			if( cam->getFOVy() < Ogre::Degree( 90 ) ) cam->setFOVy( cam->getFOVy() + Ogre::Degree( 5 ) );
		}

		environmentNormal.normalise();
		
		double yaw = ( lookLeft->getTimeFactoredAnalog() - lookRight->getTimeFactoredAnalog() ) * 30 * stableFactor;
		desiredOrientation = desiredOrientation * Quaternion( Degree( -yaw ), Vector3::UNIT_Z );
		body->setOrientation( desiredOrientation );

		double pitch = ( lookUp->getTimeFactoredAnalog() - lookDown->getTimeFactoredAnalog() ) * 30 * stableFactor;
		if( absolutePitch + pitch > 85 ) pitch = 85 - absolutePitch;
		if( absolutePitch + pitch < -85 ) pitch = -85 - absolutePitch;
		absolutePitch += pitch;
		eyeNode->pitch( Degree( pitch ) );
		
		desiredVel = desiredOrientation.xAxis() * ( moveRight->getMaximumAnalog() - moveLeft->getMaximumAnalog() ) +
					 desiredOrientation.yAxis() * ( moveBackward->getMaximumAnalog() - moveForward->getMaximumAnalog() );
		desiredVel.normalise();
		desiredVel *= speed;
		desiredVel.y = 0;
		desiredVel += Vector3::UNIT_Y * jump->getFired() * jumpAmount;

		Vector3 currentVel = body->getLinearVelocity();

		if( currentVel.y <= -20 ) fallTime += mec::gTime->getDelta();
		else 
		{
			if( fallTime && ( currentVel.y >= -7 ) )
			{
				blood = 0;
				fallTime = 0;
				return;
			}
		}
		if( fallTime >= 1 )
		{
			blood = 0;
			fallTime = 0;
			return;
		}

		if( !canReachGround ) 
		{
			desiredVel /= speed;
			desiredVel += environmentNormal * 2 * environmentNormal.y;
			desiredVel.y = 0;
			desiredVel *= speed;
		}

		if( crouch->getFired() && !crouching )
		{
			crouching = true;
			setupGeometry( true );
			if( canReachGround ) body->addForce( Vector3( 0, - mass * jumpAmount / 2, 0 ) );
		}

		if( crouch->getCeased() && crouching )
		{
			crouching = false;
			setupGeometry( false );
		}

		if( !environmentNormal.length() )
		{
			double length = jumpDirVel.length();

			if( length )
			{
				desiredVel += 0.6 * jumpDirVel;
				desiredVel *= 0.6;
			}
			else 
			{
				jumpDirVel = desiredVel.length() ? desiredVel : Vector3( 0.0001, 0, 0.0001 );
				jumpDirVel.y = 0;
			}
		}
		else
		{
			if( crouching ) 
			{
				desiredVel *= 0.4;
				desiredVel.y = 0;

				if( eyeNode->getPosition().z < - height / 2 + 0.7 ) eyeNode->translate( 0, 0, mec::gTime->getDelta() * 4 );
			}
			else
			{
				if( eyeNode->getPosition().z > - height / 2 + 0.2 ) eyeNode->translate( 0, 0, -mec::gTime->getDelta() * 4 );
			}

			jumpDirVel = Vector3::ZERO;
		}

		if( jumpingUp )
		{
			if( currentVel.y <= 0 ) jumpingUp = false;
		}
		if( desiredVel.y > 0 ) jumpingUp = true;

		if( !desiredVel.length() )
		{
			if( touchingGround && !jumpingUp )
			{
				body->setAffectedByGravity( false );
				body->setLinearVelocity( Vector3::ZERO );
				body->setAngularVelocity( Vector3::ZERO );
			}
			else body->setAffectedByGravity( true );
		}
		else 
		{
			body->setAffectedByGravity( true );
			Vector3 force = mass * Vector3( ( desiredVel.x - currentVel.x ) * mec::gTime->getDelta() / impulse, 
											min( ( double )desiredVel.y, jumpAmount ) , 
											( desiredVel.z - currentVel.z ) * mec::gTime->getDelta() / impulse );
			body->addForce( force );
		}

		touchingGround = false;
		canReachGround = false;
		environmentNormal = Vector3::ZERO;

		hitByLow = false;
		hitByMedium = false;
		hitByHigh = false;

		if( currentWeapon ) currentWeapon->update( this );
	}

	FlyCam::FlyCam()
	{
		addType( "flycam" );
	}

	void FlyCam::reference()
	{
		moveLeft = mec::gInput->getActionSignalGroup( "moveLeft" );
		moveRight = mec::gInput->getActionSignalGroup( "moveRight" );
		moveForward = mec::gInput->getActionSignalGroup( "moveForward" );
		moveBackward = mec::gInput->getActionSignalGroup( "moveBackward" );
		lookLeft = mec::gInput->getActionSignalGroup( "lookLeft" );
		lookRight = mec::gInput->getActionSignalGroup( "lookRight" );
		lookUp = mec::gInput->getActionSignalGroup( "lookUp" );
		lookDown = mec::gInput->getActionSignalGroup( "lookDown" );
	}

	void FlyCam::save( mec::Serial* outSerial )
	{
		outSerial->setAttribute( "position", StringConverter::toString( node->getPosition() ) );
		outSerial->setAttribute( "orientation", StringConverter::toString( node->getOrientation() ) );
	}

	bool FlyCam::setup( mec::Serial* inSerial )
	{
		SceneManager* sm = mec::gGraphics->getSceneManager();

		node = sm->getRootSceneNode()->createChildSceneNode( "flycam" );
		node->attachObject( sm->getCamera( "camera" ) );
		node->setPosition( StringConverter::parseVector3( inSerial->getAttribute( "position" ) ) );
		node->setOrientation( StringConverter::parseQuaternion( inSerial->getAttribute( "orientation" ) ) );

		return true;
	}

	void FlyCam::shutdown()
	{
		mec::gGraphics->getSceneManager()->destroySceneNode( "flycam" );
	}

	void FlyCam::update()
	{
		node->translate( ( moveRight->getMaximumAnalog() - moveLeft->getMaximumAnalog() ) * 50 * mec::gTime->getDelta(),
			             0,
			             ( moveBackward->getMaximumAnalog() - moveForward->getMaximumAnalog() ) * 50 * mec::gTime->getDelta(), 
						 Node::TS_LOCAL );
		node->yaw( Degree( lookLeft->getTimeFactoredAnalog() - lookRight->getTimeFactoredAnalog() ) * 30, Node::TS_WORLD );
		node->pitch( Degree( lookUp->getTimeFactoredAnalog() - lookDown->getTimeFactoredAnalog() ) * 30, Node::TS_LOCAL );
	}
}